propeller_design_tools (PDT)
============================
---
**Work in progress / incomplete documentation (all current documentation is below, as well as a step-by-step typical 
usage/workflow scenario)**

Note: The graphical user interface for this should be functional for Linux now as well!

---
Description
===========
Python 3.7 package that provides exactly what it sounds like by automating usage of the GPL-licensed 
CLI-utilities XFOIL and XROTOR.

Both utilities are published by professor Mark Drela (MIT).

- XFOIL: for arbitrary 2D airfoil analysis and design
- XROTOR: for arbitrary propeller design schemes and analysis

---
Troubleshooting PDT Errors / Crashes (GUI and/or Scripting)
===========================================================
So I just quit my job as a systems engineer to become a software developer, so hopefully this package starts to get
more frequent updates and support from me. If you encounter what you think is an error coming from this code, I would
love to know about it, please submit an issue to:

[LINK TO GITHUB ISSUES](https://github.com/helloDestroyerOfWorlds/propeller_design_tools/issues)

and I will likely see it quickly and will at least answer you, if not just fix it and push the update. Please only do
this though once you are sure the error is coming from the PDT code.

If you think you have discovered a legitimate issue, before submitting it please also read through the list below to
see if any of the listed things could fix your error.

1. XFOIL or XROTOR executables: Errors might be coming from XFOIL or XROTOR in actuality rather than this source code
    - Currently PDT implements these as executables installed when pip installing this package
    - This means if you are having pip install errors and are on Linux it might be caused by that, actually, if you are
      having any errors, even in the GUI, and you are on Linux it might be caused by me not understanding until now much
      about Linux. If this is the case, and you think you know what is causing the error, please feel free to reach out
      to me via the issues link above.
    - Additionally, I just implemented a fix to automatically install all required dependencies when pip installing this
      package. This should fix many issues for Linux users, if not all the current ones. Note that if you want to use
      the
      GUI, read the next point.
    - Still submit an issue using the link above if this is the case -> in theory I can detect most of these

2. Did not install GUI version: To enable the ability to use the graphical-user-interface (GUI) you must:

   `pip install propeller_design_tools[gui]` - or - `pip install --upgrade propeller_design_tools[gui]`

3. You're asking for an "impossible" propeller: XROTOR will error in this case.
    - Impossible in terms of the pure physics of what you're asking for (asking for too much thrust / power essentially)
    - PDT has no way to know if your propeller is impossible, so I'm not sure how I could ever wrap this error in an
      appropriate python error wrapper even.
    - I highly recommend to use only the "constant" blade cl distribution target option
        - I have not had time yet to test any of the other options much
        - Other options are disabled in the GUI until I can test them fully
        - Other options in the future will be a "linear" option as well as "file" option (which essentially allows the
        - user to define an arbitrary, discreet function for the blade Cl distribution).
    - It is also highly recommended to use only the 'grad' or 'pot' solver options until you know you have the code at
      least generating an output without failing.
        - I would only then recommend moving on to using the 'vrtx' solver. This solver is (in my opinion) the superior
          of
          the 3 options. Though it takes longest to solve by far, it seems to be the only one that reliably produces
          reasonable chord lengths near the hub. I think the 'vrtx' solver is somewhat less stable than the other two as
          well, so if you think you're having errors with PDT try to switch solvers.

4. You're asking for XFOIL to analyze your airfoil at too low or too high of an angle-of-attack (AoA)
    - I have attempted to remove the need for the user to enter AoA, but my algorithm for this is far from perfect.
    - If you think this is your error, please submit it via issue link above.

5. You're trying to open the STL file generated by PDT in SolidWorks: this is a known issue currently being worked on.
   - I've no idea why this happens, it seems to be a purely SolidWorks issue, the STL file seems to open just fine in
     everything else I have tried.
   - Try opening the file and re-saving it (I suggest MS Paint, I think the new version can save STLs).
   - Try opening the file and re-saving it after having very slightly changed it.
   - If you encounter this error but think you have figured it out or have useful information about it please submit an
     issue via the link above.

Thank you all that is my TED talk.

---
Purpose
=======
PDT seeks to provide the user a set of python3 utilities that can be used for arbitrary scripting efforts to automate
usage of both XFOIL and XROTOR while implementing its own unique python3-native algorithms to maintain local
input files, meta files, databases, and results files and weave everything together for the user in a simple,
meaningful way to aid in the initial / investigatory stage of well-behaved propeller designs (free-flow and non-ducted
assumptions).

---
Getting Started
===============
Installation
------------
`pip install propeller_design_tools` - or - `pip install propeller_design_tools[gui]`

Update
------
`pip install --upgrade propeller_design_tools` - or - `pip install --upgrade propeller_design_tools[gui]`

General Operation
-----------------
`import propeller_design_tools as pdt`

PDT operates on two different "database" directories, defined by the user with:

`pdt.set_airfoil_database(path: str)`
and
`pdt.set_propeller_database(path: str)`

**The user must set these two directories at the top of every script right after the imports**

*The airfoil directory will be used to store any foil / XFOIL-related support files, and the propeller directory
will be used similarly to store any propeller / XROTOR-related support files.*

---
Example Scripts / Workflow
--------------------------
At a high-level, the current concept for the PDT workflow is as follows (after pip-installing the PDT package):

0. Try out the (currently extremely buggy and incomplete) user interface!
   [example0_user_interface.py](
   https://github.com/helloDestroyerOfWorlds/propeller_design_tools/blob/master/tests/example0_user_interface.py)

   ![ex0-1](./tests/ex0-1.png)
   ![ex0-2](./tests/ex0-2.png)
   ![ex0-3](./tests/ex0-3.png)


1. Obtain normalized airfoil coordinate files from
[UIUC Database](https://m-selig.ae.illinois.edu/ads/coord_database.html)
and save these files into the "airfoil_database" directory (GUI users can just use the built-in tool to download)


2. Use PDT to run XFOIL across ranges of Reynolds Numbers in order to populate database data for the desired foil 
sections -> see 
[example1_airfoil_analysis.py](
   https://github.com/helloDestroyerOfWorlds/propeller_design_tools/blob/master/tests/example1_airfoil_analysis.py
   )

   ![ex1-1](./tests/ex1-1.png)
   ![ex1-2](./tests/ex1-2.png)


3. Once the required 2D airfoil data is generated, PDT can then be used to automatically generate all the required 2D 
   foil definition parameters required by XROTOR (these "station parameters" are essentially what allow XROTOR to model 
   the performance of well-behaved, arbitrarily-lofted blade geometries) -> see
[example2_radialstation_creation.py](
   https://github.com/helloDestroyerOfWorlds/propeller_design_tools/blob/master/tests/example2_radialstation_creation.py
   )

   ![ex2-1](./tests/ex2-1.png)
   
**But this step is also automated & displayed by PDT when the using the builtin PDT propeller creation function (or the 
GUI)** -> see
[example3_prop_creation.py](
   https://github.com/helloDestroyerOfWorlds/propeller_design_tools/blob/master/tests/example3_prop_creation.py
   )

   ![ex3-1](./tests/ex3-1.png)

NOTE: It is highly recommended to first run XROTOR using either the 'grad' or the 'pot' vortex formulation in order to 
get your design "tweaked in" -> these are much faster than the (more accurate) 'vrtx' formulation, which you 
can then move on to.

   ![ex3-2](./tests/ex3-2.png)


4. PDT's `Propeller()` object instances can generate 3D geometry files including profile xyz coordinate listings and 
   .stl 3D geometry files -> see
[example4_stl_generation.py](
   https://github.com/helloDestroyerOfWorlds/propeller_design_tools/blob/master/tests/example4_stl_generation.py
   )

   ![ex4-1](./tests/ex4-1.png)


5. Analyze a given `Propeller()` instance across a sweep of operating points -> see
   [example5_prop_analysis.py](
   https://github.com/helloDestroyerOfWorlds/propeller_design_tools/blob/master/tests/example5_prop_analysis.py
   )

   ![ex5-1](./tests/ex5-1.png)


6. **WIP** Prop optimization
    - Will be a grid-search style generic optimizer for "optimal" prop design generation by means of minimizing overall
      energy usage.
    - Optionally takes into account different propeller operating points via the ability to define the propeller's
      "duty-cycle"
        - This feature is coming soon!

---
XFOIL, XROTOR and other MIT Aero-Modeling Executables
-----------------------------------------------------
In order to utilize any PDT functionality that depends on
running XFOIL, the "xfoil.exe" executable file needs to be
in the user-set "airfoil_database" location. *Current pip
installations include the xfoil.exe file in the foil_database,
there should theoretically be no need to download it manually,
even if the user changes the foil_database location.*

[XFOIL executable and docs](https://web.mit.edu/drela/Public/web/xfoil/)

Likewise, in order to utilize any PDT functionality that
depends on running XROTOR, the "xrotor.exe" executable file
needs to be in the user-set "propeller_database" location.
*Current pip installations include the xrotor.exe file in the
default prop_database, there should theoretically be no need to
download it manually, even if the user changes the prop_database
location.*

[CROTOR executable and docs](http://www.esotec.org/sw/crotor.html#download)
*(this is actually a link to "CROTOR", which I find is
the easiest way to obtain a windows-executable of XROTOR)*

[actual XROTOR docs](https://web.mit.edu/drela/Public/web/xrotor/xrotor_doc.txt)

Both XFOIL and XROTOR contain much, much more utility than PDT currently has integrated. They are part of an incredible
set of aerodynamic (or hydrodynamic) modeling command-line-interface (CLI) utilities from MIT. The ones that I
personally know of include:

1. XFOIL - for arbitrary 2D airfoil analysis and design
2. XROTOR - for arbitrary propeller design schemes and analysis
3. AVL - for arbitrary lifting planform analysis
4. QPROP - another CLI utility for propeller design and analysis
5. QMIL - for arbitrary windmill designs (and more)

Combined, the above represents an amazing fluid-dynamics modeling capability that is completely free, and fast. While
it's unlikely they will ever get as accurate as true computational-fluid-dynamics (CFD) software, they still offer
great value for being free, and are several orders of magnitude faster than true CFD software. Personally, I have found
them to be completely sufficient tools for all stages of engineering design, but then again I have only ever worked in
prototyping engineering.